#!/usr/bin/env python
from subprocess import check_output, check_call, CalledProcessError
import argparse
import json
import logging
import os
import re
import sys


def which(prog, paths=None, default=''):
    """Locate executable.

    :param prog: program to find
    :type prog: str
    :param paths: if not None then we use this value instead of PATH to look
        for the executable.
    :type paths: str | None
    :param default: default value to return if not found
    :type default: str | None | T

    :return: absolute path to the program on success, found by searching for an
      executable in the directories listed in the environment variable PATH
      or default value if not found
    :rtype: str | None | T
    """
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    def possible_names(fpath):
        names = [fpath]
        if sys.platform == 'win32':
            names.extend([fpath + ext for ext in
                          os.environ.get('PATHEXT', '').split(';')])
        return names

    fpath, fname = os.path.split(prog)
    if fpath:
        # Full path given, check if executable
        for progname in possible_names(prog):
            if is_exe(progname):
                return progname
    else:
        # Check for all directories listed in $PATH
        if paths is None:
            paths = os.environ["PATH"]

        for pathdir in paths.split(os.pathsep):
            exe_file = os.path.join(pathdir, prog)
            for progname in possible_names(exe_file):
                if is_exe(progname):
                    return progname

    # Not found.
    return default


SUPPORTED_LIBRARY_TYPES = {'static', 'static-pic', 'relocatable'}


def decode_library_types(arg):
    library_types = arg.split(',')
    library_type_set = set(library_types)

    # Make sure that all requested library types are supported
    unsupported = library_type_set - SUPPORTED_LIBRARY_TYPES
    if unsupported:
        raise ValueError('Unsupported library types: {}'
                         .format(', '.join(sorted(unsupported))))

    # Make sure that the given list of library types contains no double entries
    if len(library_types) != len(library_type_set):
        raise ValueError('Library types cannot be requested twice')

    # Return the list, and not the set, to keep the original ordering
    return library_types


class Config(object):
    def __init__(self, args=None):
        self.object_dir = os.path.abspath(os.getcwd())
        self.source_dir = os.path.abspath(os.path.dirname(sys.argv[0]))
        self.json_cache = os.path.join(self.object_dir, 'setup.json')
        if args is not None:
            self.load_cache = args.load_cache
        else:
            self.load_cache = True
        self.cache_loaded = False
        self.init_data(args)

    @classmethod
    def add_arguments(cls, parser):
        parser.add_argument('--jobs', '-j',
                            help='gprbuild parallelism', default='0')
        parser.add_argument('--reconfigure',
                            help='ignore previous setup results',
                            dest='load_cache',
                            default=True,
                            action="store_false")
        parser.add_argument('--target',
                            help='target',
                            default=None)
        parser.add_argument('--prefix',
                            help='installation prefix',
                            default='auto')
        parser.add_argument('--integrated',
                            help='installation in platform specific subdir',
                            default=False,
                            action="store_true")
        parser.add_argument('--library-types',
                            default=['static', 'static-pic', 'relocatable'],
                            type=decode_library_types,
                            help='Comma-separated list of library types to'
                                 ' build. By default, build all of them.')

    def init_data(self, args):
        if self.load_cache and os.path.isfile(self.json_cache):
            with open(self.json_cache, 'r') as fd:
                try:
                    data = json.load(fd)
                except Exception:
                    data = {}
            # Check that the setup is not changed
            if (data.get('source_dir', '') == self.source_dir and
                    data.get('object_dir', '') == self.object_dir):
                self.data = data
                self.cache_loaded = True
                return

        self.data = {}
        self.data['source_dir'] = self.source_dir
        self.data['object_dir'] = self.object_dir

        if args is not None:
            if args.target is None:
                self.data['target'] = self.run('gcc', '-dumpmachine',
                                               grab=True)
            else:
                self.data['target'] = args.target

            self.data['jobs'] = args.jobs
            self.data['integrated'] = args.integrated
            self.data['canonical_target'] = self.run(
                'gprconfig', '--config=ada',
                '--target=%s' % self.data['target'],
                '--mi-show-compilers',
                grab=r' 1 normalized_target:(\S*)')

            default_prefix = os.path.dirname(
                os.path.dirname(
                    self.run('gprconfig', '--config=ada',
                             '--target=%s' % self.data['target'],
                             '--mi-show-compilers',
                             grab=r' 1 path:(.*)')))
            if args.prefix == 'auto':
                self.data['prefix'] = default_prefix
            else:
                self.data['prefix'] = args.prefix

            # The first element in library_types list define the default type
            # of library that will be used. Do not rely on the default set in
            # the project file.
            self.data['library_types'] = args.library_types

    def set_data(self, name, value, sub=None):
        if sub is None:
            self.data[name] = value
        else:
            if sub not in self.data:
                self.data[sub] = {}
            self.data[sub][name] = value

    def save_data(self):
        with open(self.json_cache, 'w') as fd:
            json.dump(self.data, fd)

    def gprcmd(self, cmd, project, *args, **kwargs):
        cmd = list(cmd) + list(args)

        # Handle out of source dir builds
        if self.source_dir != self.object_dir:
            cmd.append('-P%s' % os.path.join(self.source_dir, project))
            cmd.append('--relocate-build-tree')
        else:
            cmd.append('-P%s' % project)

        # Always pass --target so that 'Target value is enforced.
        cmd.append('--target=%s' % self.data['canonical_target'])

        # Pass common variables values computed during configure
        # step and stored in setup.json
        if 'gprbuild' in self.data:
            for name, value in self.data['gprbuild'].items():
                cmd.append('-X%s=%s' % (name, value))

        # Additional scenario variables coming usually from variants
        if 'gpr_vars' in kwargs:
            for name, value in kwargs['gpr_vars'].items():
                cmd.append('-X%s=%s' % (name, value))

        return self.run(*cmd, **kwargs)

    @property
    def prefix(self):
        prefix = self.data['prefix']
        # In integrated mode always install in a subdirectory which is
        # target specific
        if self.data['integrated']:
            prefix = os.path.join(prefix, self.data['canonical_target'])
        return prefix

    def gprbuild(self, project, *args, **kwargs):
        cmd = ['gprbuild', '-j%s' % self.data['jobs'], '-p']
        return self.gprcmd(cmd, project, *args, **kwargs)

    def gprclean(self, project, *args, **kwargs):
        cmd = ['gprclean', '-q']
        return self.gprcmd(cmd, project, *args, **kwargs)

    def gprinstall(self, project, *args, **kwargs):
        # Sources are shared between all variants and put in
        # include/<project_name> where project_name is the base name of
        # project file.
        cmd = ['gprinstall', '-p', '-f',
               '--prefix=%s' % self.prefix,
               '--sources-subdir=include/%s' % project[:-4]]

        return self.gprcmd(cmd, project, *args, **kwargs)

    def gpruninstall(self, project, *args, **kwargs):
        cmd = ['gprinstall', '-p', '-f', '--prefix=%s' % self.prefix]
        cmd.append('--uninstall')
        return self.gprcmd(cmd, project, *args, **kwargs)

    def run(self, *args, **kwargs):
        grab = kwargs.get('grab', False)

        cmd = list(args)

        cmd[0] = which(cmd[0])
        assert cmd[0], "cannot find program: %s" % args[0]

        if grab:
            output = check_output(cmd).decode('utf-8').strip()
            if not isinstance(grab, bool):
                output = re.findall(grab, output)[0]
            return output
        else:
            logging.info('Launch: %s' % " ".join(cmd))
            return check_call(cmd)


class SetupApp(object):

    def __init__(self):
        pass

    def variants(self, config, cmd):
        return [([], {})]

    def build(self, args):
        config = Config(args)
        if not config.cache_loaded:
            self.update_config(config, args)
            config.save_data()

        for gpr_args, gpr_vars in self.variants(config, 'build'):
            config.gprbuild(
                self.project,
                *(gpr_args + args.gpr_opts + ['-gargs']),
                gpr_vars=gpr_vars)
        return 0

    def clean(self, args):
        config = Config()
        if not config.cache_loaded:
            logging.info('nothing to clean')
            return 0

        for gpr_args, gpr_vars in self.variants(config, 'clean'):
            config.gprclean(self.project,
                            *(gpr_args + args.gpr_opts),
                            gpr_vars=gpr_vars)

    def install(self, args):
        config = Config()
        if not config.cache_loaded:
            logging.info('nothing to install')
            return 0
        if args.prefix is not None:
            config.set_data('prefix', args.prefix)

        logging.info('%-26s %s',
                     'Installation directory', config.data['prefix'])
        for gpr_args, gpr_vars in self.variants(config, 'install'):
            config.gprinstall(self.project,
                              *(gpr_args + args.gpr_opts),
                              gpr_vars=gpr_vars)

    def uninstall(self, args):
        config = Config()
        if not config.cache_loaded:
            logging.info('nothing to uninstall')
            return 0
        if args.prefix is not None:
            config.set_data('prefix', args.prefix)

        config.gpruninstall(self.project, *args.gpr_opts)

    def create(self):
        self.main = argparse.ArgumentParser(description=self.description)
        self.parser = self.main.add_subparsers(
            title='commands',
            description='available commands (do ./setup.py CMD --help '
                        'for help on command CMD)')

        # Build command
        self.build_cmd = self.parser.add_parser('build',
                                                help='build %s' % self.name)
        self.build_cmd.add_argument(
            '--gpr-opts',
            nargs=argparse.REMAINDER,
            default=[],
            help="pass remaining arguments to gprbuild")
        Config.add_arguments(self.build_cmd)
        self.build_cmd.set_defaults(command=self.build)

        # Clean command
        self.clean_cmd = self.parser.add_parser(
            'clean',
            help='clean %s' % self.name)
        self.clean_cmd.add_argument(
            '--gpr-opts',
            default=[],
            nargs=argparse.REMAINDER,
            help="pass remaining arguments to gprclean")
        self.clean_cmd.set_defaults(command=self.clean)

        # Install command
        self.install_cmd = self.parser.add_parser(
            'install',
            help='install %s' % self.name)
        self.install_cmd.add_argument(
                '--gpr-opts',
                default=[],
                nargs=argparse.REMAINDER,
                help="pass remaining arguments to gprinstall")
        self.install_cmd.add_argument('--prefix',
                                      help='installation prefix',
                                      default=None)
        self.install_cmd.set_defaults(command=self.install)

        # Uninstall command
        self.uninstall_cmd = self.parser.add_parser(
            'uninstall',
            help='uninstall %s' % self.name)
        self.uninstall_cmd.add_argument(
            '--gpr-opts',
            default=[],
            nargs=argparse.REMAINDER,
            help="pass remaining arguments to gpruninstall")
        self.uninstall_cmd.add_argument('--prefix',
                                        help='un-installation prefix',
                                        default=None)
        self.uninstall_cmd.set_defaults(command=self.uninstall)
        logging.basicConfig(level=logging.DEBUG, format='%(message)s')

    def run(self):
        self.create()
        args = self.main.parse_args()
        try:
            return args.command(args)
        except CalledProcessError as e:
            logging.error('process failed with status: %s', e.returncode)
            return 1
        except AssertionError as e:
            logging.error('requirement missing: %s', e)
            return 1



import os

from e3.fs import cp
from e3.testsuite.result import TestStatus

from drivers import gprbuild, GNATcollTestDriver
from drivers.valgrind import check_call_valgrind


class BasicTestDriver(GNATcollTestDriver):
    """Default GNATcoll testsuite driver.

    In order to declare a test:

    1- Create a directory with a test.yaml inside
    2- Add test sources in that directory
    3- Add a main called test.adb that use support/test_assert.ads package.
    4- Do not put test.gpr there, it breaks the test, if you need a project
       file for testing, name it something else.
    5- If you need additional files for you test, list them in test.yaml:
       data:
           - "your_file1"
           - "your_file2"
    """

    def add_test(self, dag):
        """Declare test workflow.

        The workflow is the following::

            build --> check status

        :param dag: tree of test fragment to amend
        :type dag: e3.collection.dag.DAG
        """
        self.add_fragment(dag, "build")
        self.add_fragment(dag, "check_run", after=["build"])

        if "test_exe" not in self.test_env:
            self.test_env["test_exe"] = "obj/test"

    def build(self, previous_values, slot):
        """Build fragment."""
        if self.test_env.get("no-coverage"):
            gpr_project_path = self.env.gnatcoll_prod_gpr_dir
        else:
            gpr_project_path = self.env.gnatcoll_gpr_dir
        return gprbuild(
            self, gcov=self.env.gcov, gpr_project_path=gpr_project_path
        )

    def check_run(self, previous_values, slot):
        """Check status fragment."""
        if not previous_values["build"]:
            return

        for data in self.test_env.get("data", []):
            cp(
                os.path.join(self.test_env["test_dir"], data),
                self.test_env["working_dir"],
                recursive=True,
            )

        process = check_call_valgrind(
            self,
            [
                os.path.join(
                    self.test_env["working_dir"], self.test_env["test_exe"]
                )
            ],
        )
        if "<=== TEST PASSED ===>" not in process.out:
            self.result.set_status(TestStatus.FAIL)
        else:
            self.result.set_status(TestStatus.PASS)
        self.push_result()



#! /usr/bin/env python3
from e3.testsuite import Testsuite
from e3.testsuite.driver import TestDriver
from e3.testsuite.result import TestStatus
from e3.testsuite.process import check_call
from e3.env import Env
from e3.fs import mkdir, cp, ls
import logging
import sys
import os

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))


class DefaultDriver(TestDriver):
    @property
    def project_file(self):
        result = os.path.join(self.test_env["test_dir"], "test.gpr")
        if not os.path.isfile(result):
            result = os.path.join(self.env.root_dir, "support", "test.gpr")
        return result

    @property
    def build_dir(self):
        return self.test_env["working_dir"]

    @property
    def test_source_dir(self):
        return self.test_env["test_dir"]

    @property
    def support_source_dir(self):
        return os.path.join(self.env.root_dir, "support")

    def build(self, prev, slot):
        self.logger = logging.getLogger(f"test.{self.test_env['test_name']}")

        env = {
            "TEST_SOURCES": self.test_source_dir,
            "SUPPORT_SOURCES": self.support_source_dir,
        }

        mkdir(self.build_dir)
        py_files = ls(os.path.join(self.test_source_dir, "*.py"))
        if py_files:
            cp(py_files, self.build_dir)
        check_call(
            self,
            ["gprbuild", "-P", self.project_file, "--relocate-build-tree", "-p"],
            cwd=self.build_dir,
            timeout=300,
            env=env,
            ignore_environ=False,
        )

    def run(self, prev, slot):
        if self.result.status == TestStatus.ERROR:  # means that status was not set
            test_py = os.path.join(self.build_dir, "test.py")
            if os.path.isfile(test_py):
                self.test_process = check_call(
                    self, [sys.executable, "./test.py"], cwd=self.build_dir, timeout=60
                )
            else:
                self.test_process = check_call(
                    self,
                    [os.path.join(self.build_dir, "obj", "test")]
                    + self.test_env.get("test_args", []),
                    timeout=60,
                    cwd=self.build_dir,
                )

    def analyze(self, prev, slot):
        if self.result.status == TestStatus.ERROR:  # means that status was not set
            if "<=== TEST PASSED ===>" not in self.test_process.out:
                self.result.set_status(TestStatus.FAIL)
            else:
                self.result.set_status(TestStatus.PASS)
            self.push_result()

    def add_test(self, dag):
        self.add_fragment(dag, "build")
        self.add_fragment(dag, "run", after=["build"])
        self.add_fragment(dag, "analyze", after=["run"])


class GNATCOLLPython3Test(Testsuite):
    """Testsuite for the bc(1) calculator."""

    test_driver_map = {"default": DefaultDriver}
    default_driver = "default"
    tests_subdir = "tests"


if __name__ == "__main__":
    Env.add_search_path("GPR_PROJECT_PATH", os.path.join(ROOT_DIR, "support"))
    assert (
        "ADA_PYTHON_HOME" in os.environ
    ), "ADA_PYTHON_HOME should point to Python distrib used to build the binding"
    sys.exit(GNATCOLLPython3Test().testsuite_main())



from e3.testsuite.process import check_call


def check_call_valgrind(driver, cmd, test_name=None, result=None, **kwargs):
    """
    Wrapper for `e3.testsuite.process` that runs the process under Valgrind if
    this is a Valgrind-checked testsuite run. The process exit status will be
    2 if Valgrind finds memory issues.
    """
    if driver.env.valgrind:
        cmd = [
            "valgrind",
            "-q",
            "--error-exitcode=2",
            "--leak-check=full",
        ] + cmd
    return check_call(driver, cmd, test_name, result, **kwargs)



#!/usr/bin/env python

import re
import os
import logging
import sys

from e3.fs import mkdir, ls, find
from e3.os.process import Run
from e3.testsuite import Testsuite

from drivers import make_gnatcoll, TESTSUITE_ROOT_DIR
from drivers.basic import BasicTestDriver


class MyTestsuite(Testsuite):

    tests_subdir = "tests"
    test_driver_map = {
        "default": BasicTestDriver,
    }
    default_driver = "default"

    def add_options(self, parser):
        parser.add_argument(
            "--gcov",
            help="compute testsuite coverage of gnatcoll",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--valgrind",
            help="check memory usage with Valgrind (memcheck tool)",
            action="store_true",
        )
        parser.add_argument(
            "--recompile",
            help="recompile production version of gnatcoll for testing",
            default=None,
            action="store_const",
            const="PROD",
        )
        parser.add_argument(
            "--debug",
            help="recompile debug version of gnatcoll for testing",
            dest="recompile",
            action="store_const",
            const="DEBUG",
        )

    def set_up(self):
        super().set_up()

        self.env.gcov = self.main.args.gcov
        self.env.valgrind = self.main.args.valgrind
        if self.main.args.gcov:
            work_dir = os.path.join(TESTSUITE_ROOT_DIR, "gcov")
            gpr_dir, src_dir, obj_dir = make_gnatcoll(work_dir, "DEBUG", gcov=True)
            self.env.gnatcoll_gpr_dir = gpr_dir
            self.env.gnatcoll_src_dir = src_dir
            self.env.gnatcoll_obj_dir = obj_dir
        else:
            self.env.gnatcoll_gpr_dir = None

        recompile_mode = self.main.args.recompile
        if recompile_mode:
            work_dir = os.path.join(TESTSUITE_ROOT_DIR, recompile_mode.lower())
            gpr_dir, _, _ = make_gnatcoll(
                work_dir, recompile_mode, gcov=False
            )
            self.env.gnatcoll_prod_gpr_dir = gpr_dir
            if self.env.gnatcoll_gpr_dir is None:
                self.env.gnatcoll_gpr_dir = gpr_dir
        else:
            self.env.gnatcoll_prod_gpr_dir = None

    def tear_down(self):
        if self.main.args.gcov:
            wd = TESTSUITE_ROOT_DIR

            # We need to call gcov on gcda present both in gnatcoll itself and
            # tests (for generics coverage).
            gcda_files = find(
                os.path.join(self.env.gnatcoll_obj_dir), "*.gcda"
            ) + find(os.path.join(self.env.working_dir), "*.gcda")
            mkdir(os.path.join(wd, "gcov", "results"))
            gcr = os.path.join(wd, "gcov", "results")
            Run(["gcov"] + gcda_files, cwd=os.path.join(wd, "gcov", "results"))
            total_sources = 0
            total_covered = 0

            for source_file in ls(
                os.path.join(self.env.gnatcoll_src_dir, "*", "*")
            ):
                base_file = os.path.basename(source_file)
                if not os.path.isfile(os.path.join(gcr, base_file + ".gcov")):
                    total = 1
                    covered = 0
                    with open(source_file) as fd:
                        total = len(
                            [
                                line
                                for line in fd
                                if line.strip() and not re.match(r" *--", line)
                            ]
                        )
                else:
                    with open(os.path.join(gcr, base_file + ".gcov")) as fd:
                        total = 0
                        covered = 0
                        for line in fd:
                            if re.match(r" *-:", line):
                                pass
                            elif re.match(r" *[#=]{5}:", line):
                                total += 1
                            else:
                                total += 1
                                covered += 1
                total_sources += total
                total_covered += covered

                logging.info(
                    "%6.2f %% %8d/%-8d %s",
                    float(covered) * 100.0 / float(total),
                    covered,
                    total,
                    os.path.basename(source_file),
                )

            logging.info(
                "%6.2f %% %8d/%-8d %s",
                float(total_covered) * 100.0 / float(total_sources),
                total_covered,
                total_sources,
                "TOTAL",
            )
        super().tear_down()


if __name__ == "__main__":
    sys.exit(MyTestsuite().testsuite_main())



import logging
import os
import sys
import traceback

from e3.fs import mkdir
from e3.os.fs import df
from e3.os.process import Run
from e3.testsuite.driver import TestDriver
from e3.testsuite.process import check_call
from e3.testsuite.result import TestStatus


# Root directory of respectively the testsuite and the gnatcoll
# repository.
TESTSUITE_ROOT_DIR = os.path.dirname(
    os.path.dirname(os.path.abspath(__file__))
)
GNATCOLL_ROOT_DIR = os.path.dirname(TESTSUITE_ROOT_DIR)

PROJECT = {
    "gmp": "gnatcoll_gmp.gpr",
    "iconv": "gnatcoll_iconv.gpr",
    "lzma": "gnatcoll_lzma.gpr",
    "omp": "gnatcoll_omp.gpr",
    "zlib": "gnatcoll_zlib.gpr",
    "cpp": "gnatcoll_cpp.gpr",
}


def make_gnatcoll(work_dir, build_mode, gcov=False):
    """Build gnatcoll core with or without gcov instrumentation.

    :param work_dir: working directory. gnatcoll is built in `build` subdir
        and installed in `install` subdir
    :type work_dir: str
    :param build_mode: build mode to use in order to build the gnatcoll
        project
    :param gcov: if True, build it with gcov instrumentation
    :type gcov: bool
    :return: a triplet (project path, source path, object path)
    :rtype: (str, str, str)
    :raise AssertError: in case compilation of installation fails
    """
    params = f"({build_mode}, gcov={gcov})"

    # Create build tree structure
    build_dir = os.path.join(work_dir, "build")
    install_dir = os.path.join(work_dir, "install")
    mkdir(build_dir)
    mkdir(install_dir)

    # Compute make invocation
    for binding in PROJECT.keys():
        logging.info(f"Compiling gnatcoll {binding} {params}")
        setup = os.path.join(GNATCOLL_ROOT_DIR, binding, "setup.py")
        obj_dir = os.path.join(build_dir, binding)
        mkdir(obj_dir)

        build_cmd = [sys.executable, setup, "build", "--library-types=static"]
        install_cmd = [
            sys.executable,
            setup,
            "install",
            "--prefix",
            install_dir,
        ]

        if gcov:
            build_cmd += [
                "--gpr-opts",
                "-cargs",
                "-fprofile-arcs",
                "-ftest-coverage",
                "-largs",
                "-lgcov",
                "-gargs",
                f"-XBUILD={build_mode}",
            ]
        else:
            build_cmd += ["--gpr-opts", f"-XBUILD={build_mode}"]

        # Build & Install
        p = Run(build_cmd, cwd=obj_dir)
        assert p.status == 0, "gnatcoll %s build failed:\n%s" % (
            binding,
            p.out,
        )
        logging.debug("build:\n%s", p.out)

        p = Run(install_cmd, cwd=obj_dir)
        assert p.status == 0, "gnatcoll %s installation failed:\n%s" % (
            binding,
            p.out,
        )
        logging.debug("install:\n%s", p.out)

    return (
        os.path.join(install_dir, "share", "gpr"),
        os.path.join(install_dir, "include"),
        build_dir,
    )


def gprbuild(
    driver,
    project_file=None,
    cwd=None,
    gcov=False,
    scenario=None,
    gpr_project_path=None,
    **kwargs
):
    """Launch gprbuild.

    :param project_file: project file to compile. If None, we looks first for
        a test.gpr in the test dir and otherwise fallback on the common
        test.gpr project of the support subdir of the testsuite.
    :type project_file: str
    :param cwd: directory in which to run gprbuild. If None the gprbuild build
        is run in the default working dir for the test.
    :type cwd: str | None
    :param gcov: if True link with gcov libraries
    :type gcov: bool
    :param scenario: scenario variable values
    :type scenario: dict
    :param gpr_project_path: if not None prepent this value to GPR_PROJECT_PATH
    :type gpr_project_path: None | str
    :param kwargs: additional keyword arguements are passed to
        e3.testsuite.process.check_call function
    :return: True on successful completion
    :rtype: bool
    """
    if scenario is None:
        scenario = {}

    if cwd is None:
        cwd = driver.test_env["working_dir"]
    mkdir(cwd)

    if project_file is None:
        project_file = os.path.join(driver.test_env["test_dir"], "test.gpr")
        if not os.path.isfile(project_file):
            project_file = os.path.join(cwd, "test.gpr")
            with open(
                os.path.join(TESTSUITE_ROOT_DIR, "support", "test.gpr"), "r"
            ) as fd:
                content = fd.read()
            with open(project_file, "w") as fd:
                for component in driver.test_env.get("components", []):
                    fd.write('with "%s";\n' % PROJECT[component])
                fd.write(content)
            scenario["TEST_SOURCES"] = driver.test_env["test_dir"]
    scenario["SUPPORT_SOURCES"] = os.path.join(TESTSUITE_ROOT_DIR, "support")

    gprbuild_cmd = [
        "gprbuild",
        "--relocate-build-tree",
        "-p",
        "-P",
        project_file,
    ]
    for k, v in scenario.items():
        gprbuild_cmd.append("-X%s=%s" % (k, v))
    if gcov:
        gprbuild_cmd += [
            "-largs",
            "-lgcov",
            "-cargs",
            "-fprofile-arcs",
            "-ftest-coverage",
            "-g",
        ]

    # Adjust process environment
    env = None
    if gpr_project_path:
        new_gpr_path = gpr_project_path
        if "GPR_PROJECT_PATH" in os.environ:
            new_gpr_path += os.path.pathsep + os.environ["GPR_PROJECT_PATH"]
        env = {"GPR_PROJECT_PATH": new_gpr_path}

    check_call(
        driver, gprbuild_cmd, cwd=cwd, env=env, ignore_environ=False, **kwargs
    )
    # If we get there it means the build succeeded.
    return True


class GNATcollTestDriver(TestDriver):
    """Abstract class to share some common facilities."""

    def should_skip(self):
        """Handle of 'skip' in test.yaml.

        :return: None if the test should not be skipped, a TestStatus
            otherwise.
        :rtype: None | TestStatus
        """
        if "skip" in self.test_env:
            eval_env = {
                "env": self.env,
                "test_env": self.test_env,
                "disk_space": lambda: df(self.env.working_dir),
            }

            for status, expr in self.test_env["skip"]:
                try:
                    if eval(expr, eval_env):
                        return TestStatus[status]
                except Exception:
                    logging.error(traceback.format_exc())
                    return TestStatus.ERROR
        return None



#!/usr/bin/env python
import logging
import sys
import re
import os
import json
import shutil

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from setup_support import SetupApp, Config

PYTHON_DATA_SCRIPT = """
import json
from sysconfig import get_config_vars, get_path, get_path_names
if "platinclude" in get_path_names():
    plat_include_key = "platinclude"
else:
    plat_include_key = "include"
result = {'config_vars': get_config_vars(),
          'python_inc': get_path("include"),
          'python_inc_plat': get_path(plat_include_key),
          'prefix': get_path("data")}
print(json.dumps(result))
"""


def fetch_python_config(config):
    logging.info("Fetch Python information...")
    python_output = config.run(
        config.data["python_exec"], "-c", PYTHON_DATA_SCRIPT, grab=True
    )
    python_data = json.loads(python_output)
    config_vars = python_data["config_vars"]
    python_version = config_vars["VERSION"]
    python_ldversion = config_vars.get("LDVERSION", python_version)
    logging.info("  %-24s %s", "Python version:", python_version)

    # Current python location
    current_prefix = python_data["prefix"]

    # Fetch prefix during the build process. Some paths of interest might
    # still reference a location used during the Python build process.
    build_prefix = (
        [sys.prefix]
        + re.findall(r"'--prefix=([^']+)'", config_vars.get("CONFIG_ARGS", ""))
    )[-1]

    def relocate(path):
        if os.path.isabs(path):
            rel_path = os.path.relpath(path, build_prefix)
            if not rel_path.startswith(os.pardir):
                # If the input path is relative to the original build
                # directory, replace build prefix by the current one.
                return os.path.join(current_prefix, rel_path)
            else:
                # Otherwise, return it unchanged
                return path
        else:
            # The input path is relative so assume it's relative to the current
            # python prefix.
            return os.path.join(current_prefix, path)

    # Retrieve cflags, and linker flags
    static_dir = relocate(config_vars.get("LIBPL", "libs"))
    logging.info("  %-24s %s", "Static dir", static_dir)

    shared_dir = relocate(config_vars.get("LIBDIR", "."))
    logging.info("  %-24s %s", "Shared dir", shared_dir)

    # Add first relocated include dirs followed by non-relocated version.
    # Indeed, when using venv and maybe virtualenv, includes are not copied.
    cflags = "-I" + relocate(python_data["python_inc"])
    if python_data["python_inc"] != python_data["python_inc_plat"]:
        cflags += " -I" + relocate(python_data["python_inc_plat"])
        cflags += " -I" + python_data["python_inc_plat"]
    cflags += " -I" + python_data["python_inc"]

    logging.info("  %-24s %s", "CFLAGS", cflags)

    if python_version.startswith("3"):
        # In python 3.x MODLIBS seems to drag in too many libraries
        python_libs = [
            config_vars[v]
            for v in ("LIBS", "SYSLIBS")
            if v in config_vars and config_vars[v]
        ]
    else:
        python_libs = [
            config_vars[v]
            for v in ("LIBS", "SYSLIBS", "MODLIBS")
            if v in config_vars and config_vars[v]
        ]
    python_libs = " ".join(python_libs)

    python_shared_libs = "-L%s -lpython%s %s" % (
        shared_dir,
        python_ldversion,
        python_libs,
    )
    python_static_libs = python_libs
    libpython_a = os.path.join(
        static_dir, config_vars.get("LIBRARY", "libpython%s.a" % python_version)
    )

    if os.path.isfile(libpython_a):
        config.set_data("GNATCOLL_PYTHON_STATIC_LIB", libpython_a, sub="gprbuild")
    else:
        logging.info("static python library not found")

    if sys.platform.startswith("linux"):
        # On Linux platform, even when linking with the static libpython,
        # symbols not used by the application itself should be exported so
        # that shared library present in Python can use the Python C API.
        python_static_libs += " -export-dynamic"
        python_shared_libs += " -export-dynamic"

    logging.info("  %-24s %s", "Shared linker flags", python_shared_libs)
    logging.info("  %-24s %s", "Static linker flags", python_static_libs)

    # User does not have the choice between linking with static libpython
    # and shared libpython. If --enable-shared or --enable-framework was
    # passed to Python's configure during Python build, then we should
    # link with the shared libpython, otherwise with the static one.
    # Indeed otherwise some C modules might not work as expected or even
    # crash. On Windows always link with shared version of libpython
    # (if the static is present, this is just an indirection to the shared)
    if (
        "--enable-shared" in config_vars.get("CONFIG_ARGS", "")
        or "--enable-framework" in config_vars.get("CONFIG_ARGS", "")
        or sys.platform.startswith("win")
    ):
        logging.info("Force link to shared python library")
        config.set_data("GNATCOLL_PYTHON_LIBS", python_shared_libs, sub="gprbuild")
        config.set_data("GNATCOLL_LIBPYTHON_KIND", "shared", sub="gprbuild")
    else:
        logging.info("Force link to static python library")
        config.set_data("GNATCOLL_PYTHON_LIBS", python_static_libs, sub="gprbuild")
        config.set_data("GNATCOLL_LIBPYTHON_KIND", "static", sub="gprbuild")
    config.set_data("GNATCOLL_PYTHON_CFLAGS", cflags, sub="gprbuild")


class GNATCollPython(SetupApp):
    name = "gnatcoll_python"
    project = "gnatcoll_python.gpr"
    description = "GNATColl Python bindings"

    def create(self):
        super(GNATCollPython, self).create()
        self.build_cmd.add_argument(
            "--python-exec",
            help="set python executable location",
            metavar="PATH",
            default=sys.executable,
        )
        self.build_cmd.add_argument(
            "--debug",
            help="build project in debug mode",
            action="store_true",
            default=False,
        )

    def update_config(self, config, args):
        # Fetch python information
        config.set_data("python_exec", args.python_exec)
        fetch_python_config(config)

        logging.info(
            "%-26s %s", "Libraries kind", ", ".join(config.data["library_types"])
        )

        # Set library version
        with open(
            os.path.join(config.source_dir, "..", "version_information"), "r"
        ) as fd:
            version = fd.read().strip()
        config.set_data("GNATCOLL_VERSION", version, sub="gprbuild")
        logging.info("%-26s %s", "Version", version)

        # Set build mode
        config.set_data("BUILD", "DEBUG" if args.debug else "PROD", sub="gprbuild")
        logging.info("%-26s %s", "Build mode", config.data["gprbuild"]["BUILD"])

        # Set GNATCOLL_OS
        if "darwin" in config.data["canonical_target"]:
            gnatcoll_os = "osx"
        elif "windows" in config.data["canonical_target"]:
            gnatcoll_os = "windows"
        else:
            # Assume this is an Unix system
            gnatcoll_os = "unix"
        config.set_data("GNATCOLL_OS", gnatcoll_os, sub="gprbuild")

    def variants(self, config, cmd):
        result = []
        for library_type in config.data["library_types"]:
            gpr_vars = {
                "LIBRARY_TYPE": library_type,
                "XMLADA_BUILD": library_type,
                "GPR_BUILD": library_type,
            }
            if cmd == "install":
                result.append(
                    (
                        ["--build-name=%s" % library_type, "--build-var=LIBRARY_TYPE"],
                        gpr_vars,
                    )
                )
            else:
                result.append(([], gpr_vars))
        return result

    def install(self, args):
        config = Config()
        has_static_python = "GNATCOLL_PYTHON_STATIC_LIB" in config.data["gprbuild"]

        if not has_static_python:
            super(GNATCollPython, self).install(args)
        else:
            python_la = config.data["gprbuild"]["GNATCOLL_PYTHON_STATIC_LIB"]
            prefix = args.prefix
            if prefix is None:
                prefix = config.data["prefix"]
            rel_target = os.path.join(
                "..", "..", "lib", "gnatcoll_python.static", os.path.basename(python_la)
            )
            abs_target = os.path.join(
                prefix, "lib", "gnatcoll_python.static", os.path.basename(python_la)
            )

            shutil.copy(config.json_cache, config.json_cache + ".backup")
            try:
                # Temporary change the configuration to set a relative path to the
                # static Python library.
                config.set_data(
                    "GNATCOLL_PYTHON_STATIC_LIB", rel_target, sub="gprbuild"
                )
                config.save_data()

                # Perform the installation
                super(GNATCollPython, self).install(args)

                # Copy over the libpython*.a
                logging.info("Copy static libpython into target lib")
                if not os.path.isdir(os.path.dirname(abs_target)):
                    os.mkdir(abs_target)
                result = shutil.copy(python_la, abs_target)
                logging.info(f"Python static lib in {result}")
            finally:
                # Restore the configuration cache
                shutil.copy(config.json_cache + ".backup", config.json_cache)
                os.unlink(config.json_cache + ".backup")


if __name__ == "__main__":
    app = GNATCollPython()
    sys.exit(app.run())
